"""Initializes the .fmu directory."""

from pathlib import Path
from textwrap import dedent
from typing import Any, Final

from fmu.datamodels.fmu_results.global_configuration import GlobalConfiguration

from ._fmu_dir import ProjectFMUDirectory, UserFMUDirectory
from ._logging import null_logger
from .models.project_config import ProjectConfig

logger: Final = null_logger(__name__)

_README = dedent("""\
    This directory contains static configuration data for your FMU project.

    You should *not* manually modify files within this directory. Doing so may
    result in erroneous behavior or erroneous data in your FMU project.

    Changes to data stored within this directory must happen through the FMU
    Settings application.

    Run `fmu-settings` to do this.
""")

_USER_README = dedent("""\
    This directory contains static data and configuration elements used by some
    components in FMU. It may also contains sensitive access tokens that should not be
    shared with others.

    You should *not* manually modify files within this directory. Doing so may
    result in erroneous behavior by some FMU components.

    Changes to data stored within this directory must happen through the FMU
    Settings application.

    Run `fmu-settings` to do this.
""")


def _create_fmu_directory(base_path: Path) -> None:
    """Creates the .fmu directory.

    Args:
        base_path: Base directory where .fmu should be created

    Raises:
        FileNotFoundError: If base_path doesn't exist
        FileExistsError: If .fmu exists
    """
    logger.debug(f"Creating .fmu directory in '{base_path}'")

    if not base_path.exists():
        raise FileNotFoundError(
            f"Base path '{base_path}' does not exist. Expected the root "
            "directory of an FMU project."
        )

    fmu_dir = base_path / ".fmu"
    if fmu_dir.exists():
        if fmu_dir.is_dir():
            raise FileExistsError(f"{fmu_dir} already exists")
        raise FileExistsError(f"{fmu_dir} exists but is not a directory")

    fmu_dir.mkdir()
    logger.debug(f"Created .fmu directory at '{fmu_dir}'")


def init_fmu_directory(
    base_path: str | Path,
    config_data: ProjectConfig | dict[str, Any] | None = None,
    global_config: GlobalConfiguration | None = None,
) -> ProjectFMUDirectory:
    """Creates and initializes a .fmu directory.

    Also initializes a configuration file if configuration data is provided through the
    function.

    Args:
        base_path: Directory where .fmu should be created.
        config_data: Optional ProjectConfig instance or dictionary with configuration
          data.
        global_config: Optional GlobaConfiguration instance with existing global config
          data.

    Returns:
        Instance of FMUDirectory

    Raises:
        FileExistsError: If .fmu exists
        FileNotFoundError: If base_path doesn't exist
        PermissionError: If the user lacks permission to create directories
        ValidationError: If config_data fails validationg
    """
    logger.debug("Initializing .fmu directory")
    base_path = Path(base_path)

    _create_fmu_directory(base_path)

    fmu_dir = ProjectFMUDirectory(base_path)
    fmu_dir.write_text_file("README", _README)

    fmu_dir.config.reset()
    if config_data:
        if isinstance(config_data, ProjectConfig):
            config_data = config_data.model_dump()
        fmu_dir.update_config(config_data)

    if global_config:
        for key, value in global_config.model_dump().items():
            fmu_dir.set_config_value(key, value)

    logger.info(f"Successfully initialized .fmu directory at '{fmu_dir}'")
    return fmu_dir


def init_user_fmu_directory() -> UserFMUDirectory:
    """Creates and initializes a user's $HOME/.fmu directory.

    Returns:
        Instance of FMUDirectory

    Raises:
        FileExistsError: If .fmu exists
        FileNotFoundError: If base_path doesn't exist
        PermissionError: If the user lacks permission to create directories
        ValidationError: If config_data fails validationg
    """
    logger.debug("Initializing .fmu directory")

    _create_fmu_directory(Path.home())

    fmu_dir = UserFMUDirectory()
    fmu_dir.write_text_file("README", _USER_README)

    fmu_dir.config.reset()
    logger.debug(f"Successfully initialized .fmu directory at '{fmu_dir}'")
    return fmu_dir



"""The generic configuration file in a .fmu directory."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Self, TypeVar

from pydantic import ValidationError

from fmu.settings._logging import null_logger
from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.models.user_config import UserConfig
from fmu.settings.types import ResettableBaseModel  # noqa: TC001

from .pydantic_resource_manager import PydanticResourceManager

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        FMUDirectoryBase,
        ProjectFMUDirectory,
        UserFMUDirectory,
    )

logger: Final = null_logger(__name__)

T = TypeVar("T", bound=ResettableBaseModel)


class ConfigManager(PydanticResourceManager[T]):
    """Manages the .fmu configuration file."""

    def __init__(self: Self, fmu_dir: FMUDirectoryBase, config: type[T]) -> None:
        """Initializes the Config resource manager."""
        super().__init__(fmu_dir, config)

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the config file."""
        return Path("config.json")

    def _get_dot_notation_key(
        self: Self, config_dict: dict[str, Any], key: str, default: Any = None
    ) -> Any:
        """Sets the value to a dot-notation key.

        Args:
            config_dict: The configuration dictionary we are modifying (by reference)
            key: The key to set
            default: Value to return if key is not found. Default None

        Returns:
            The value or default
        """
        parts = key.split(".")
        value = config_dict
        for part in parts:
            if isinstance(value, dict) and part in value:
                value = value[part]
            else:
                return default

        return value

    def get(self: Self, key: str, default: Any = None) -> Any:
        """Gets a configuration value by key.

        Supports dot notation for nested values (e.g., "foo.bar")

        Args:
            key: The configuration key
            default: Value to return if key is not found. Default None

        Returns:
            The configuration value or default
        """
        try:
            config = self.load()

            if "." in key:
                return self._get_dot_notation_key(config.model_dump(), key, default)

            if hasattr(config, key):
                return getattr(config, key)

            config_dict = config.model_dump()
            return config_dict.get(key, default)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Resource file for '{self.__class__.__name__}' not found "
                f"at: '{self.path}' when getting key {key}"
            ) from e

    def _set_dot_notation_key(
        self: Self, config_dict: dict[str, Any], key: str, value: Any
    ) -> None:
        """Sets the value to a dot-notation key.

        Args:
            config_dict: The configuration dictionary we are modifying (by reference)
            key: The key to set
            value: The value to set
        """
        parts = key.split(".")
        target = config_dict

        for part in parts[:-1]:
            if part not in target or not isinstance(target[part], dict):
                target[part] = {}
            target = target[part]

        target[parts[-1]] = value

    def set(self: Self, key: str, value: Any) -> None:
        """Sets a configuration value by key.

        Args:
            key: The configuration key
            value: The value to set

        Raises:
            FileNotFoundError: If config file doesn't exist
            ValueError: If the updated config is invalid
        """
        try:
            config = self.load()
            config_dict = config.model_dump()

            if "." in key:
                self._set_dot_notation_key(config_dict, key, value)
            else:
                config_dict[key] = value

            updated_config = config.model_validate(config_dict)
            self.save(updated_config)
        except ValidationError as e:
            raise ValueError(
                f"Invalid value set for '{self.__class__.__name__}' with "
                f"key '{key}', value '{value}': '{e}"
            ) from e
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Resource file for '{self.__class__.__name__}' not found "
                f"at: '{self.path}' when setting key {key}"
            ) from e

    def update(self: Self, updates: dict[str, Any]) -> T:
        """Updates multiple configuration values at once.

        Args:
            updates: Dictionary of key-value pairs to update

        Returns:
            The updated Config object

        Raises:
            FileNotFoundError: If config file doesn't exist
            ValueError: If the updates config is invalid
        """
        try:
            config = self.load()
            config_dict = config.model_dump()

            flat_updates = {k: v for k, v in updates.items() if "." not in k}
            config_dict.update(flat_updates)

            for key, value in updates.items():
                if "." in key:
                    self._set_dot_notation_key(config_dict, key, value)

                updated_config = config.model_validate(config_dict)
            self.save(updated_config)
        except ValidationError as e:
            raise ValueError(
                f"Invalid value set for '{self.__class__.__name__}' with "
                f"updates '{updates}': '{e}"
            ) from e
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Resource file for '{self.__class__.__name__}' not found "
                f"at: '{self.path}' when setting updates {updates}"
            ) from e

        return updated_config

    def reset(self: Self) -> T:
        """Resets the configuration to defaults.

        Returns:
            The new default config object
        """
        config = self.model_class.reset()
        self.save(config)
        return config


class ProjectConfigManager(ConfigManager[ProjectConfig]):
    """Manages the .fmu configuration file in a project."""

    def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
        """Initializes the ProjectConfig resource manager."""
        super().__init__(fmu_dir, ProjectConfig)


class UserConfigManager(ConfigManager[UserConfig]):
    """Manages the .fmu configuration file in a user's home directory."""

    def __init__(self: Self, fmu_dir: UserFMUDirectory) -> None:
        """Initializes the UserConfig resource manager."""
        super().__init__(fmu_dir, UserConfig)



"""The model for config.json."""

from __future__ import annotations

from datetime import UTC, datetime
from pathlib import Path
from typing import Annotated, Self

import annotated_types
from pydantic import (
    AwareDatetime,
    BaseModel,
    SecretStr,
    field_serializer,
    field_validator,
)

from fmu.settings import __version__
from fmu.settings.types import ResettableBaseModel, VersionStr  # noqa: TC001

RecentProjectDirectories = Annotated[list[Path], annotated_types.Len(0, 5)]


class UserAPIKeys(BaseModel):
    """Known API keys stored in a user config."""

    smda_subscription: SecretStr | None = None

    @field_serializer("smda_subscription", when_used="json")
    def dump_secret(self, v: SecretStr | None) -> str | None:
        """Write the secret string value when serializing to json."""
        if v is None:
            return None
        return v.get_secret_value()


class UserConfig(ResettableBaseModel):
    """The configuration file in a $HOME/.fmu directory.

    Stored as config.json.
    """

    version: VersionStr
    created_at: AwareDatetime
    user_api_keys: UserAPIKeys
    recent_project_directories: RecentProjectDirectories

    @classmethod
    def reset(cls: type[Self]) -> Self:
        """Resets the model to an initial state."""
        return cls(
            version=__version__,
            created_at=datetime.now(UTC),
            user_api_keys=UserAPIKeys(),
            recent_project_directories=[],
        )

    @field_validator("recent_project_directories", mode="before")
    @classmethod
    def ensure_unique(cls, recent_projects: list[Path]) -> list[Path]:
        """Ensures that recent_project_directories contains unique entries."""
        if len(recent_projects) != len(set(recent_projects)):
            raise ValueError("recent_project_directories must contain unique entries")
        return recent_projects

    def obfuscate_secrets(self: Self) -> Self:
        """Returns a copy of the model with obfuscated secrets.

        If an API Key is:

            key: SecretStr = SecretStr("secret")

        we may want to serialize it to JSON as:

            {key:"********"}

        so that we do not serialize the actual value of the secret when, for example,
        returning the user configuration from an API.
        """
        config_dict = self.model_dump()
        # Overwrite secret keys with obfuscated keys
        for k, v in config_dict["user_api_keys"].items():
            if v is not None:
                # Convert SecretStr("*********") to "*********"
                config_dict["user_api_keys"][k] = str(v)
        return self.model_validate(config_dict)



"""Tests specific to the user configuration.

Because the user config and the project config share the same base class a great deal of
the functionality is already tested in test_project_config.py. This test module is more
sparse to handle things specific to the user configuration.
"""

import json

import pytest
from pydantic import SecretStr

from fmu.settings._fmu_dir import UserFMUDirectory


def test_set_smda_subscription_key_validates(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests that setting a subscription key is validated by Pydantic."""
    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is None
    with pytest.raises(ValueError, match="Invalid value set"):
        user_fmu_dir.set_config_value("user_api_keys.smda_subscription", 123)
    assert config.user_api_keys.smda_subscription is None

    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", "secret")

    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is not None
    assert config.user_api_keys.smda_subscription.get_secret_value() == "secret"

    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", None)
    config = user_fmu_dir.config.load()


def test_set_smda_subscription_key_writes_to_disk(
    user_fmu_dir: UserFMUDirectory,
) -> None:
    """Tests that setting a subscription key updates to disk."""
    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", "secret")
    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is not None
    assert config.user_api_keys.smda_subscription.get_secret_value() == "secret"
    assert config.user_api_keys.smda_subscription == SecretStr("secret")
    assert user_fmu_dir.get_config_value(
        "user_api_keys.smda_subscription"
    ) == SecretStr("secret")
    assert (
        user_fmu_dir.get_config_value(
            "user_api_keys.smda_subscription"
        ).get_secret_value()
        == "secret"
    )

    with open(user_fmu_dir.config.path, encoding="utf-8") as f:
        config_dict = json.loads(f.read())
    assert config_dict["user_api_keys"]["smda_subscription"] == "secret"


def test_obfuscate_secrets(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests the obfuscate_secrets() method on the UserConfig model."""
    user_fmu_dir.set_config_value("user_api_keys.smda_subscription", "secret")
    config = user_fmu_dir.config.load()
    assert config.user_api_keys.smda_subscription is not None
    assert config.user_api_keys.smda_subscription.get_secret_value() == "secret"
    assert config.user_api_keys.smda_subscription == SecretStr("secret")

    config_dump = config.model_dump()
    assert config_dump["user_api_keys"]["smda_subscription"] == SecretStr("secret")
    config_json = config.model_dump_json()
    assert '{"smda_subscription":"secret"}' in config_json

    secret_config = config.obfuscate_secrets()
    secret_config_dump = secret_config.model_dump()
    assert secret_config_dump["user_api_keys"]["smda_subscription"] == SecretStr(
        "**********"
    )
    secret_config_json = secret_config.model_dump_json()
    assert '{"smda_subscription":"**********"}' in secret_config_json



"""Contains the base class used for interacting with resources."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Generic, Self, TypeVar

from pydantic import BaseModel, ValidationError

if TYPE_CHECKING:
    from pathlib import Path

    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import FMUDirectoryBase

T = TypeVar("T", bound=BaseModel)


class PydanticResourceManager(Generic[T]):
    """Base class for managing resources represented by Pydantic models."""

    def __init__(self: Self, fmu_dir: FMUDirectoryBase, model_class: type[T]) -> None:
        """Initializes the resource manager.

        Args:
            fmu_dir: The FMUDirectory instance
            model_class: The Pydantic model class this manager handles
        """
        self.fmu_dir = fmu_dir
        self.model_class = model_class
        self._cache: T | None = None

    @property
    def relative_path(self: Self) -> Path:
        """Returns the path to the resource file _inside_ the .fmu directory.

        Must be implemented by subclasses.
        """
        raise NotImplementedError

    @property
    def path(self: Self) -> Path:
        """Returns the full path to the resource file."""
        return self.fmu_dir.get_file_path(self.relative_path)

    @property
    def exists(self: Self) -> bool:
        """Returns whether or not the resource exists."""
        return self.path.exists()

    def load(self: Self, force: bool = False, store_cache: bool = True) -> T:
        """Loads the resource from disk and validates it as a Pydantic model.

        Args:
            force: Force a re-read even if the file is already cached.
            store_cache: Whether or not to cache the loaded model internally. This is
                best used with 'force=True' because if a model is already stored in
                _cache it will be returned without re-loading. Default True.

        Returns:
            Validated Pydantic model

        Raises:
            ValueError: If the resource file is missing or data does not match the
            model schema
        """
        if self._cache is None or force:
            if not self.exists:
                raise FileNotFoundError(
                    f"Resource file for '{self.__class__.__name__}' not found "
                    f"at: '{self.path}'"
                )

            try:
                content = self.fmu_dir.read_text_file(self.relative_path)
                data = json.loads(content)
                validated_model = self.model_class.model_validate(data)
                if store_cache:
                    self._cache = validated_model
                else:
                    return validated_model
            except ValidationError as e:
                raise ValueError(
                    f"Invalid content in resource file for '{self.__class__.__name__}: "
                    f"'{e}"
                ) from e
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"Invalid JSON in resource file for '{self.__class__.__name__}': "
                    f"'{e}'"
                ) from e

        return self._cache

    def save(self: Self, model: T) -> None:
        """Save the Pydantic model to disk.

        Args:
            model: Validated Pydantic model instance
        """
        self.fmu_dir._lock.ensure_can_write()
        json_data = model.model_dump_json(by_alias=True, indent=2)
        self.fmu_dir.write_text_file(self.relative_path, json_data)
        self._cache = model



"""Tests for the ProjectFMUDirectory class."""

import json
from pathlib import Path
from unittest.mock import patch

import pytest
from pytest import MonkeyPatch

from fmu.settings import __version__, find_nearest_fmu_directory, get_fmu_directory
from fmu.settings._fmu_dir import ProjectFMUDirectory, UserFMUDirectory
from fmu.settings._resources.lock_manager import LockManager


def test_init_existing_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests initializing an ProjectFMUDirectory on an existing .fmu directory."""
    fmu = ProjectFMUDirectory(fmu_dir.base_path)
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path


def test_get_fmu_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests initializing an ProjectFMUDirectory via get_fmu_directory."""
    fmu = get_fmu_directory(fmu_dir.base_path)
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path


def test_find_nearest_fmu_directory(
    monkeypatch: MonkeyPatch, fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests initializing an ProjectFMUDirectory via find_nearest_fmu_directory."""
    subdir = fmu_dir.path / "subdir"
    subdir.mkdir()
    subdir2 = fmu_dir.path / "subdir2"
    subdir2.mkdir()
    subsubdir = subdir / "subsubdir"
    subsubdir.mkdir()

    fmu = find_nearest_fmu_directory(str(subsubdir))
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path

    monkeypatch.chdir(fmu_dir.base_path)
    fmu = find_nearest_fmu_directory()
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path

    monkeypatch.chdir(subdir2)
    fmu = find_nearest_fmu_directory()
    assert fmu.path == fmu_dir.path
    assert fmu.base_path == fmu_dir.base_path


def test_init_on_missing_directory(tmp_path: Path) -> None:
    """Tests initializing with a missing directory raises."""
    with pytest.raises(
        FileNotFoundError, match=f"No .fmu directory found at {tmp_path}"
    ):
        ProjectFMUDirectory(tmp_path)


def test_init_when_fmu_is_not_a_directory(tmp_path: Path) -> None:
    """Tests initialized on a .fmu non-directory raises."""
    (tmp_path / ".fmu").touch()
    with pytest.raises(
        FileExistsError, match=f".fmu exists at {tmp_path} but is not a directory"
    ):
        ProjectFMUDirectory(tmp_path)


def test_find_fmu_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests find_fmu_directory method on nested children."""
    child = fmu_dir.base_path / "child"
    grand_child = child / "grandchild"
    grand_child.mkdir(parents=True)

    found_dir = ProjectFMUDirectory.find_fmu_directory(grand_child)
    assert found_dir == fmu_dir.path


def test_find_fmu_directory_not_found(tmp_path: Path) -> None:
    """Tests find_fmu_directory() returns None if no .fmu found."""
    empty_dir = tmp_path / "empty"
    empty_dir.mkdir()

    found_dir = ProjectFMUDirectory.find_fmu_directory(empty_dir)
    assert found_dir is None


def test_find_nearest(fmu_dir: ProjectFMUDirectory) -> None:
    """Test find_nearest factory method."""
    subdir = fmu_dir.base_path / "subdir"
    subdir.mkdir()

    fmu = ProjectFMUDirectory.find_nearest(subdir)
    assert fmu.path == fmu_dir.path


def test_find_nearest_not_found(tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
    """Test find_nearest raises FileNotFoundError when not found."""
    monkeypatch.chdir(tmp_path)
    with pytest.raises(
        FileNotFoundError, match=f"No .fmu directory found at or above {tmp_path}"
    ):
        ProjectFMUDirectory.find_nearest()
    with pytest.raises(
        FileNotFoundError, match=f"No .fmu directory found at or above {tmp_path}"
    ):
        ProjectFMUDirectory.find_nearest(tmp_path)


def test_get_config_value(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests get_config_value retrieves correctly from the config."""
    assert fmu_dir.get_config_value("version") == __version__
    assert fmu_dir.get_config_value("created_by") == "user"


def test_set_config_value(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests set_config_value sets and writes the result."""
    fmu_dir.set_config_value("version", "200.0.0")
    with open(fmu_dir.config.path, encoding="utf-8") as f:
        config_dict = json.loads(f.read())

    assert config_dict["version"] == "200.0.0"
    assert fmu_dir.get_config_value("version") == "200.0.0"
    assert fmu_dir.config.load().version == "200.0.0"


def test_update_config(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests update_config updates and saves the config for multiple values."""
    updated_config = fmu_dir.update_config({"version": "2.0.0", "created_by": "user2"})

    assert updated_config.version == "2.0.0"
    assert updated_config.created_by == "user2"

    assert fmu_dir.config.load() is not None
    assert fmu_dir.get_config_value("version", None) == "2.0.0"
    assert fmu_dir.get_config_value("created_by", None) == "user2"

    config_file = fmu_dir.config.path
    with open(config_file, encoding="utf-8") as f:
        saved_config = json.load(f)

    assert saved_config["version"] == "2.0.0"
    assert saved_config["created_by"] == "user2"


def test_update_config_invalid_data(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that update_config raises ValidationError on bad data."""
    updates = {"version": 123}
    with pytest.raises(
        ValueError,
        match=f"Invalid value set for 'ProjectConfigManager' with updates '{updates}'",
    ):
        fmu_dir.update_config(updates)


def test_get_file_path(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests get_file_path returns correct path."""
    path = fmu_dir.get_file_path("test.txt")
    assert path == fmu_dir.path / "test.txt"


def test_file_exists(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests file_exists returns correct boolean."""
    test_file = fmu_dir.path / "exists.txt"
    test_file.touch()

    assert fmu_dir.file_exists("exists.txt") is True
    assert fmu_dir.file_exists("doesnt.txt") is False


def test_read_file(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests read_file reads bytes correctly."""
    test_file = fmu_dir.path / "bin.dat"
    test_data = b"test bin data"
    test_file.write_bytes(test_data)

    data = fmu_dir.read_file("bin.dat")
    assert data == test_data


def test_read_file_not_found(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests read_file raises FileNotFoundError for missing files."""
    with pytest.raises(FileNotFoundError, match="No such file or directory"):
        fmu_dir.read_file("not_real.txt")


def test_read_text_file(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests read_text_file reads text correctly."""
    test_file = fmu_dir.path / "text.txt"
    test_text = "test text data å"
    test_file.write_text(test_text)

    text = fmu_dir.read_text_file("text.txt")
    assert text == test_text


def test_write_text_file(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests write_text_file writes text correctly."""
    test_text = "new text data æ"
    fmu_dir.write_text_file("new_text.txt", test_text)

    file_path = fmu_dir.path / "new_text.txt"
    assert file_path.exists()
    assert file_path.read_text() == test_text


def test_write_file_creates_dir(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests write_file creates parent directories."""
    test_data = b"nested data"
    fmu_dir.write_file("nested/dir/file.dat", test_data)

    nested_dir = fmu_dir.path / "nested" / "dir"
    assert nested_dir.is_dir()

    file_path = nested_dir / "file.dat"
    assert file_path.exists()
    assert file_path.read_bytes() == test_data


def test_write_operations_raise_when_locked(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests write helpers raise when another process holds the lock."""
    lock = LockManager(fmu_dir)
    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.acquire()

    with pytest.raises(PermissionError, match="Cannot write to .fmu directory"):
        fmu_dir.write_text_file("blocked.txt", "blocked")

    with pytest.raises(PermissionError, match="Cannot write to .fmu directory"):
        fmu_dir.write_file("blocked.bin", b"blocked")

    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.release()


def test_list_files(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that list_files returns the correct files."""
    (fmu_dir.path / "file1.txt").touch()
    (fmu_dir.path / "file2.txt").touch()

    subdir = fmu_dir.path / "subdir"
    subdir.mkdir()
    (subdir / "file3.txt").touch()

    files = fmu_dir.list_files()
    filenames = [f.name for f in files]

    assert "file1.txt" in filenames
    assert "file2.txt" in filenames
    assert "config.json" in filenames

    assert "file3.txt" not in filenames

    subdir_files = fmu_dir.list_files("subdir")
    assert len(subdir_files) == 1
    assert subdir_files[0].name == "file3.txt"

    not_subdir_files = fmu_dir.list_files("not_subdir")
    assert not_subdir_files == []


def test_ensure_directory(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that ensure_directory creates directories."""
    dir_path = fmu_dir.ensure_directory("nested/test/dir")
    assert dir_path.exists()
    assert dir_path.is_dir()


def test_user_init_existing_directory(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests initializing an ProjectFMUDirectory on an existing .fmu directory."""
    with patch("pathlib.Path.home", return_value=user_fmu_dir.base_path):
        fmu = UserFMUDirectory()

    assert fmu.path == user_fmu_dir.path
    assert fmu.base_path == user_fmu_dir.base_path


def test_user_init_on_missing_directory(tmp_path: Path) -> None:
    """Tests initializing with a missing directory raises."""
    with (
        patch("pathlib.Path.home", return_value=tmp_path),
        pytest.raises(
            FileNotFoundError, match=f"No .fmu directory found at {tmp_path}"
        ),
    ):
        UserFMUDirectory()


def test_user_init_when_fmu_is_not_a_directory(tmp_path: Path) -> None:
    """Tests initialized on a .fmu non-directory raises."""
    (tmp_path / ".fmu").touch()
    with (
        patch("pathlib.Path.home", return_value=tmp_path),
        pytest.raises(
            FileExistsError, match=f".fmu exists at {tmp_path} but is not a directory"
        ),
    ):
        UserFMUDirectory()


def test_update_user_config(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests update_config updates and saves the user config for multiple values."""
    recent_dir = "/foo/bar"
    updated_config = user_fmu_dir.update_config(
        {"version": "2.0.0", "recent_project_directories": [recent_dir]}
    )

    assert updated_config.version == "2.0.0"
    assert updated_config.recent_project_directories == [Path(recent_dir)]

    assert user_fmu_dir.config.load() is not None
    assert user_fmu_dir.get_config_value("version", None) == "2.0.0"
    assert user_fmu_dir.get_config_value("recent_project_directories") == [
        Path(recent_dir)
    ]

    config_file = user_fmu_dir.config.path
    with open(config_file, encoding="utf-8") as f:
        saved_config = json.load(f)

    assert saved_config["version"] == "2.0.0"
    assert saved_config["recent_project_directories"] == [recent_dir]


def test_update_user_config_invalid_data(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests that update_config raises ValidationError on bad data."""
    updates = {"recent_project_directories": [123]}
    with pytest.raises(
        ValueError,
        match="Invalid value set for 'UserConfigManager' with updates "
        "'{'recent_project_directories':",
    ):
        user_fmu_dir.update_config(updates)


def test_update_user_config_non_unique_recent_projects(
    user_fmu_dir: UserFMUDirectory,
) -> None:
    """Tests that update_config raises on non-unique recent_project_directories."""
    updates = {"recent_project_directories": [Path("/foo/bar"), Path("/foo/bar")]}
    with pytest.raises(ValueError, match="unique entries"):
        user_fmu_dir.update_config(updates)


def test_acquire_lock_on_project_fmu(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that a lock can be acquired on the project dir."""
    fmu_dir._lock.acquire()
    assert fmu_dir._lock.is_acquired()
    assert fmu_dir._lock.exists
    assert (fmu_dir.path / ".lock").exists()


def test_acquire_lock_on_user_fmu(
    user_fmu_dir: UserFMUDirectory,
) -> None:
    """Tests that a lock can be acquired on the user dir."""
    user_fmu_dir._lock.acquire()
    assert user_fmu_dir._lock.is_acquired()
    assert user_fmu_dir._lock.exists
    assert (user_fmu_dir.path / ".lock").exists()



"""Models related to locking a .fmu directory against writes."""

from pydantic import BaseModel, Field

from fmu.settings import __version__
from fmu.settings.types import VersionStr


class LockInfo(BaseModel):
    """Represents a .fmu directory lock file."""

    pid: int
    """Process ID of the lock holder."""

    hostname: str
    """Hostname where the lock was acquired."""

    user: str
    """User who acquired the lock."""

    acquired_at: float
    """Unix timestamp when lock was acquired."""

    expires_at: float
    """Unix timestamp when lock expires."""

    version: VersionStr = Field(default=__version__)
    """The fmu-settings version acquiring the lock.

    Used for debugging."""



"""The fmu-settings package."""

try:
    from ._version import version

    __version__: str = version
except ImportError:
    __version__ = version = "0.0.0"

from ._fmu_dir import ProjectFMUDirectory, find_nearest_fmu_directory, get_fmu_directory

__all__ = ["get_fmu_directory", "ProjectFMUDirectory", "find_nearest_fmu_directory"]



"""Manages a .lock file in a .fmu/ directory."""

from __future__ import annotations

import contextlib
import os
import socket
import time
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Final, Literal, Self

from fmu.settings._logging import null_logger
from fmu.settings.models.lock_info import LockInfo

from .pydantic_resource_manager import PydanticResourceManager

if TYPE_CHECKING:
    from types import TracebackType

    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        FMUDirectoryBase,
    )

logger: Final = null_logger(__name__)

DEFAULT_LOCK_TIMEOUT: Final[int] = 1200  # 20 minutes


class LockError(Exception):
    """Raised when the lock cannot be acquired."""


class LockManager(PydanticResourceManager[LockInfo]):
    """Manages the .lock file."""

    def __init__(
        self: Self,
        fmu_dir: FMUDirectoryBase,
        timeout_seconds: int = DEFAULT_LOCK_TIMEOUT,
    ) -> None:
        """Initializes the lock manager.

        Args:
            fmu_dir: The FMUDirectory instance
            timeout_seconds: Lock expiration time in seconds (default 20 minutes)
        """
        super().__init__(fmu_dir, LockInfo)
        self._timeout_seconds = timeout_seconds
        self._acquired_at: float | None = None

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the .lock file."""
        return Path(".lock")

    def acquire(self: Self, wait: bool = False, wait_timeout: float = 5.0) -> None:
        """Acquire the lock.

        Args:
            wait: If true, wait for lock to become available
            wait_timeout: Maximum time to wait in seconds

        Raises:
            LockError: If lock cannot be acquired
        """
        if self._acquired_at is not None:
            raise LockError("Lock already acquired")

        if wait and wait_timeout <= 0:
            raise ValueError("wait_timeout must be positive")

        start_time = time.time()

        while True:
            if self.exists and self._is_stale():
                with contextlib.suppress(OSError):
                    self.path.unlink()

            if self._try_acquire():
                return

            if not wait:
                lock_info = self._safe_load()
                if lock_info:
                    raise LockError(
                        f"Lock file is held by {lock_info.user}@{lock_info.hostname} "
                        f"(PID: {lock_info.pid}). "
                        f"Expires at: {time.ctime(lock_info.expires_at)}."
                    )
                raise LockError(
                    f"Invalid lock file exists at {self.path}. "
                    "This file should be removed."
                )

            if time.time() - start_time > wait_timeout:
                raise LockError(f"Timeout waiting for lock after {wait_timeout}s")

            time.sleep(0.5)

    def _try_acquire(self: Self) -> bool:
        """Try to acquire lock. Returns True if successful.

        This method creates the lock as a temporary file first to avoid race conditions.
        Because we are operating on networked filesystems (NFS) the situation is a bit
        more complex, but os.link() is an atomic operation, so try to link the temp file
        after creating it.

        Returns:
            True if acquiring the lock succeeded
        """
        acquired_at = time.time()
        lock_info = LockInfo(
            pid=os.getpid(),
            hostname=socket.gethostname(),
            user=os.getenv("USER", "unknown"),
            acquired_at=acquired_at,
            expires_at=acquired_at + self._timeout_seconds,
        )

        temp_path = (
            self.path.parent
            / f".lock.{lock_info.hostname}.{lock_info.pid}.{uuid.uuid4().hex[:8]}"
        )
        lock_fd: int | None = None
        try:
            # O_WRONLY will allow only this process to write to the lock file.
            lock_fd = os.open(temp_path, os.O_CREAT | os.O_WRONLY | os.O_EXCL)
            try:
                json_data = lock_info.model_dump_json(by_alias=True, indent=2)
                os.write(lock_fd, json_data.encode())
                os.fsync(lock_fd)
            except Exception:
                os.close(lock_fd)
                raise

            os.close(lock_fd)
            lock_fd = None

            try:
                os.link(temp_path, self.path)
                self._cache = lock_info
                self._acquired_at = acquired_at
                return True
            except (OSError, FileExistsError):
                return False
        finally:  # Clean up temp file before we leave
            if lock_fd is not None:
                with contextlib.suppress(OSError):
                    os.close(lock_fd)
            with contextlib.suppress(OSError):
                temp_path.unlink()

    def is_locked(self: Self) -> bool:
        """Returns whether or not the lock is locked by anyone.

        This does a force load on the lock file.
        """
        lock_info = self._safe_load(force=True)
        if not lock_info:
            return False
        return time.time() < lock_info.expires_at

    def is_acquired(self: Self) -> bool:
        """Returns whether or not the lock is currently acquired by this instance."""
        if self._cache is None or self._acquired_at is None:
            return False
        return self._is_mine(self._cache) and not self._is_stale()

    def ensure_can_write(self: Self) -> None:
        """Raise PermissionError if another process currently holds the lock."""
        try:
            lock_info = self.load(force=True, store_cache=False)
        except Exception:
            lock_info = None

        if (
            self.exists
            and lock_info is not None
            and not self.is_acquired()
            and not self._is_stale(lock_info=lock_info)
        ):
            raise PermissionError(
                "Cannot write to .fmu directory because it is locked by "
                f"{lock_info.user}@{lock_info.hostname} (PID: {lock_info.pid}). "
                f"Lock expires at {time.ctime(lock_info.expires_at)}."
            )

    def refresh(self: Self) -> None:
        """Refresh/extend the lock expiration time.

        Raises:
            LockError: If we don't hold the lock or it's invalid
        """
        if not self.exists:
            raise LockError("Cannot refresh: lock file does not exist")

        lock_info = self._safe_load()
        if not lock_info or not self._is_mine(lock_info):
            raise LockError(
                "Cannot refresh: lock file is held by another process or host."
            )

        lock_info.expires_at = time.time() + self._timeout_seconds
        self.save(lock_info)

    def release(self: Self) -> None:
        """Release the lock."""
        if self.exists:
            lock_info = self._safe_load()
            if lock_info and self._is_mine(lock_info):
                with contextlib.suppress(ValueError):
                    self.path.unlink()

        self._acquired_at = None
        self._cache = None

    def save(self: Self, data: LockInfo) -> None:
        """Save the lockfile in an NFS-atomic manner.

        This overrides save() from the Pydantic resource manager.
        """
        lock_info = self._safe_load()
        if not lock_info or not self._is_mine(lock_info):
            raise LockError(
                "Failed to save lock: lock file is held by another process or host."
            )

        temp_path = Path(f"{self.path}.tmp.{uuid.uuid4().hex[:8]}")
        try:
            with open(temp_path, "w", encoding="utf-8") as f:
                f.write(data.model_dump_json(indent=2))
                f.flush()
                os.fsync(f.fileno())
            temp_path.replace(self.path)
            self._cache = data
        except Exception as e:
            with contextlib.suppress(OSError):
                temp_path.unlink()
            raise LockError(f"Failed to save lock: {e}") from e

    def _is_mine(self: Self, lock_info: LockInfo) -> bool:
        """Verifies if the calling process owns the lock."""
        return (
            lock_info.pid == os.getpid()
            and lock_info.hostname == socket.gethostname()
            and lock_info.acquired_at == self._acquired_at
        )

    def _safe_load(self: Self, force: bool = False) -> LockInfo | None:
        """Load lock info, returning None if corrupted.

        Because this file does not exist in a static state, wrap around loading it.
        """
        try:
            return self.load(force=force)
        except Exception:
            return None

    def _is_stale(self: Self, lock_info: LockInfo | None = None) -> bool:
        """Check if existing lock is stale (expired or process dead)."""
        if lock_info is None:
            lock_info = self._safe_load()

        if not lock_info:
            return True

        if time.time() > lock_info.expires_at:
            return True

        # If we aren't on the same host, we can't check the PID, so assume it's
        # not stale.
        if lock_info.hostname != socket.gethostname():
            return False

        try:
            # Doesn't actually kill, just checks if it exists
            os.kill(lock_info.pid, 0)
            return False
        except OSError:
            return True

    def __enter__(self: Self) -> Self:
        """Context manager entry."""
        self.acquire()
        return self

    def __exit__(
        self: Self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Context manager exit."""
        self.release()
        return False

    def __del__(self: Self) -> None:
        """Clean-up if garbage collected."""
        if self._acquired_at is not None:
            self.release()



"""Contains the logger that should be used throughout this package."""

import logging


def null_logger(name: str) -> logging.Logger:
    """Create and return a logger with a NullHandler.

    This function creates a logger for the specified name and attaches a
    NullHandler to it. The NullHandler prevents logging messages from being
    automatically output to the console or other default handlers. This is
    particularly useful in library modules where you want to provide the
    users of the library the flexibility to configure their own logging behavior.

    Args:
        name: The name of the logger to be created. This is typically
                    the name of the module in which the logger is
                    created (e.g., using __name__).

    Returns:
        A logger object configured with a NullHandler.

    Example:
        # In a library module
        logger = null_logger(__name__)
        logger.info("This info won't be logged to the console by default.")
    """
    logger = logging.getLogger(name)
    logger.addHandler(logging.NullHandler())
    return logger



"""Contains enumerations used in this package."""

from enum import StrEnum


class MappingType(StrEnum):
    """The discriminator used between mappings.

    Each of these types should have their own mapping class derived of some sort of
    mapping.
    """

    fault = "fault"
    stratigraphy = "stratigraphy"
    well = "well"


class RelationType(StrEnum):
    """The kind of relation this mapping represents."""

    alias = "alias"
    child_to_parent = "child_to_parent"
    equivalent = "equivalent"
    fmu_to_target = "fmu_to_target"
    predecessor_to_successor = "predecessor_to_successor"


class DataEntrySource(StrEnum):
    user = "user"
    automated = "automated"


class TargetSystem(StrEnum):
    smda = "smda"



"""Tests for LockManager."""

import json
import multiprocessing
import multiprocessing.queues
import os
import socket
import threading
import time
import uuid
from pathlib import Path
from unittest.mock import patch

import pytest
from pytest import MonkeyPatch

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.lock_manager import (
    DEFAULT_LOCK_TIMEOUT,
    LockError,
    LockManager,
)
from fmu.settings.models.lock_info import LockInfo


def test_lock_manager_instantiation(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests basic facts about the LockManager."""
    lock = LockManager(fmu_dir)

    assert lock.fmu_dir == fmu_dir
    assert lock.model_class == LockInfo
    assert lock._cache is None
    assert lock._timeout_seconds == DEFAULT_LOCK_TIMEOUT
    assert lock._acquired_at is None

    # Resource manager requires this to be implemented
    assert lock.relative_path == Path(".lock")
    assert lock.path == fmu_dir.path / lock.relative_path


def test_lock_acquire_creates_lock_file(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that acquiring a lock creates a correct lock file."""
    monkeypatch.setenv("USER", "user")

    lock = LockManager(fmu_dir)

    assert lock.exists is False
    with pytest.raises(
        FileNotFoundError, match="Resource file for 'LockManager' not found"
    ):
        lock.load()

    lock.acquire()
    assert lock.exists
    assert lock._acquired_at is not None

    lock_info = lock.load()
    assert lock._cache == lock_info
    assert lock._acquired_at == lock_info.acquired_at
    assert lock_info.pid == os.getpid()
    assert lock_info.hostname == socket.gethostname()
    assert lock_info.user == "user"
    assert lock_info.expires_at == pytest.approx(
        lock_info.acquired_at + DEFAULT_LOCK_TIMEOUT, rel=1e-05
    )


def test_lock_as_context_manager(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that acquiring a ctx manager lock creates a correct lock file."""
    monkeypatch.setenv("USER", "user")

    with LockManager(fmu_dir) as lock:
        assert lock.exists
        assert lock._acquired_at is not None

        lock_info = lock.load()
        assert lock._cache == lock_info
        assert lock._acquired_at == lock_info.acquired_at
        assert lock_info.pid == os.getpid()
        assert lock_info.hostname == socket.gethostname()
        assert lock_info.user == "user"
        assert lock_info.expires_at == pytest.approx(
            lock_info.acquired_at + DEFAULT_LOCK_TIMEOUT, rel=1e-05
        )

    assert (fmu_dir.path / ".lock").exists() is False


def test_lock_acquire_raises_if_already_acquired(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that a lock cannot be acquired twice."""
    lock = LockManager(fmu_dir)
    lock.acquire()
    assert lock.is_locked()
    assert lock.is_acquired()
    with pytest.raises(LockError, match="Lock already acquired"):
        lock.acquire()


@pytest.mark.xfail(reason="Lock is not race consistent yet")
def test_lock_acquire_race_in_threads(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that under thread race conditions one lock succeeds, one fails."""
    results = []
    errors = []

    def acquire_lock() -> None:
        """Creates, acquires, and releases a lock."""
        try:
            lock = LockManager(fmu_dir)
            lock.acquire()
            results.append("success")
        except LockError as e:
            errors.append(e)

    thread1 = threading.Thread(target=acquire_lock)
    thread2 = threading.Thread(target=acquire_lock)

    thread1.start()
    thread2.start()

    thread1.join()
    thread2.join()

    assert len(results) == 1
    assert len(errors) == 1


def _acquire_lock(
    fmu_dir: ProjectFMUDirectory,
    result_queue: multiprocessing.queues.Queue,  # type: ignore
) -> None:
    """Creates, acquires, and releases a lock.

    This is at the module level so it can be pickled by the process test.
    """
    try:
        lock = LockManager(fmu_dir)
        lock.acquire()
        result_queue.put(("success", None))
    except LockError as e:
        result_queue.put(("error", str(e)))


@pytest.mark.xfail(reason="Lock is not race consistent yet")
def test_lock_acquire_race_in_processes(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that under same process race conditions, one lock succeeds, one fails."""
    # See https://github.com/python/cpython/issues/99509 for type ignore
    result_queue = multiprocessing.Queue()  # type: ignore
    process1 = multiprocessing.Process(
        target=_acquire_lock, args=(fmu_dir, result_queue)
    )
    process2 = multiprocessing.Process(
        target=_acquire_lock, args=(fmu_dir, result_queue)
    )

    process1.start()
    process2.start()
    process1.join()
    process2.join()

    results = []
    while not result_queue.empty():
        results.append(result_queue.get())

    successes = [r for r in results if r[0] == "success"]
    errors = [r for r in results if r[0] == "error"]

    assert len(successes) == 1
    assert len(errors) == 1


def test_lock_acquire_raises_if_invalid_wait_period(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that a lock acquire requires a positive timeout."""
    lock = LockManager(fmu_dir)
    with pytest.raises(ValueError, match="wait_timeout must be positive"):
        lock.acquire(wait=True, wait_timeout=-1)
    lock.acquire(wait=True, wait_timeout=0.01)


def test_lock_acquire_over_expired_lock(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that a stale lock file will be unlinked and overwritten."""
    stale_lock = LockManager(fmu_dir, timeout_seconds=-1)  # Expired
    stale_lock.acquire()
    assert stale_lock.is_locked() is False
    assert stale_lock.is_acquired() is False
    assert stale_lock._is_stale() is True
    stale_lock_info = stale_lock.load()

    lock = LockManager(fmu_dir)
    lock.acquire()
    lock_info = lock.load()

    assert lock.is_locked()
    assert lock.is_acquired()
    assert stale_lock.path == lock.path
    assert lock_info != stale_lock_info


def test_no_wait_invalid_lock_file_exists(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that an existing, invalid .lock file raises.

    This should not really occur, but it's theoretically possible the lock file is
    overwritten to an invalid state between its stale check.
    """
    fmu_dir.write_text_file(".lock", "")
    lock = LockManager(fmu_dir)

    # This shouldn't be possible, but test that it's caught anyway
    with (
        patch.object(lock, "_is_stale", return_value=False),
        pytest.raises(LockError, match="Invalid lock file exists"),
    ):
        lock.acquire()


def test_lock_acquire_when_fresh_lock_exists_without_timeout(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that trying to acquire a fresh lock without timeout fails."""
    monkeypatch.setenv("USER", "user")

    lock = LockManager(fmu_dir)
    lock.acquire()

    bad_lock = LockManager(fmu_dir)
    with pytest.raises(LockError, match="Lock file is held by user@"):
        bad_lock.acquire()


def test_lock_with_wait_timeout_raises_after_timeout(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that trying to acquire a fresh lock with timeout succeeds."""
    lock = LockManager(fmu_dir)
    lock.acquire()

    wait_lock = LockManager(fmu_dir)
    with pytest.raises(LockError, match="Timeout waiting for lock"):
        wait_lock.acquire(wait=True, wait_timeout=0.25)


def test_lock_with_wait_timeout_succeeds_after_release(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that trying to acquire a fresh lock with timeout succeeds."""
    lock = LockManager(fmu_dir)
    lock.acquire()

    def try_acquire_second_lock() -> LockManager:
        wait_lock = LockManager(fmu_dir)
        wait_lock.acquire(wait=True, wait_timeout=1.0)
        return wait_lock

    thread = threading.Thread(target=try_acquire_second_lock)
    thread.start()

    time.sleep(0.15)
    lock.release()
    thread.join()
    assert not thread.is_alive()


def test_is_stale_expired(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests is_stale if lock has expired."""
    lock = LockManager(fmu_dir, timeout_seconds=-1)  # Expired
    lock.acquire()
    assert lock._is_stale() is True


def test_is_stale_not_expired(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests is_stale if lock has not expired."""
    lock = LockManager(fmu_dir)
    lock.acquire()
    assert lock._is_stale() is False


def test_is_stale_load_fails(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests is_stale if loading the lock file fails."""
    lock = LockManager(fmu_dir)
    lock.acquire()
    with patch.object(lock, "_safe_load", return_value=None):
        assert lock._is_stale() is True


def test_is_stale_bad_hostname(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests is_stale if lock check occurs from different host."""
    lock = LockManager(fmu_dir)
    lock.acquire()
    assert lock._is_stale() is False
    with patch(
        "fmu.settings._resources.lock_manager.socket.gethostname", return_value="foo"
    ):
        assert lock._is_stale() is False


def test_is_stale_invalid_pid(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests helper method check if lock pid does not exist."""
    lock = LockManager(fmu_dir)
    lock.acquire()
    assert lock._is_stale() is False
    assert lock._cache is not None
    lock._cache.pid = 99999999  # Should be an impossible pid
    assert lock._is_stale() is True


def test_try_acquire_succeeds_as_expected(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that try_acquire creates a correct lock file."""
    monkeypatch.setenv("USER", "user")
    time_time = 1234.5
    temp_uuid = uuid.uuid4()
    lock = LockManager(fmu_dir)
    with (
        patch("time.time", return_value=time_time),
        patch("os.getpid", return_value=123),
        patch("socket.gethostname", return_value="foo"),
        patch("pathlib.Path.unlink") as mock_unlink,
        patch("uuid.uuid4", return_value=temp_uuid),
    ):
        assert lock._try_acquire() is True
        assert lock._acquired_at == time_time
        assert lock._cache == LockInfo(
            pid=123,
            hostname="foo",
            user="user",
            acquired_at=time_time,
            expires_at=time_time + DEFAULT_LOCK_TIMEOUT,
        )
        mock_unlink.assert_called_once()

    lock_info = lock._cache
    assert lock_info is not None
    temp_lockfile = (
        fmu_dir.path / f".lock.{lock_info.hostname}.{lock_info.pid}.{temp_uuid.hex[:8]}"
    )
    # We mocked unlink(), so the temporary lockfile should still exist.
    assert temp_lockfile.exists()
    assert LockInfo.model_validate(json.loads(temp_lockfile.read_text())) == lock_info

    assert lock.path.exists()
    assert LockInfo.model_validate(json.loads(lock.path.read_text())) == lock_info


def test_try_acquire_fails_when_writing_temp_file(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that try_acquire raises when failing to write the lock file.."""
    lock = LockManager(fmu_dir)
    with (
        patch("os.write", side_effect=OSError("oops")),
        patch("os.close") as mock_close,
        patch("pathlib.Path.unlink") as mock_unlink,
        pytest.raises(OSError, match="oops"),
    ):
        lock._try_acquire()
        mock_close.assert_called_twice()
        mock_unlink.assert_called_once()


def test_try_acquire_fails_when_linking_temp_file(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests that try_acquire raises when failing to link the lock file.."""
    lock = LockManager(fmu_dir)
    with (
        patch("os.link") as mock_link,
        patch("os.close", side_effect=FileExistsError) as mock_close,
        patch("pathlib.Path.unlink") as mock_unlink,
        pytest.raises(FileExistsError),
    ):
        assert lock._try_acquire() is False
        mock_link.assert_called_once_with(lock.path)
        mock_close.assert_called_twice()
        mock_unlink.assert_called_once()


def test_is_locked_expected(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests is_locked under expected, simple conditions."""
    lock = LockManager(fmu_dir)
    assert lock.is_locked() is False
    lock.acquire()
    assert lock.is_locked() is True
    lock.release()
    assert lock.is_locked() is False


def test_is_locked_by_other_process(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests is_locked when another process has the lock."""
    lock = LockManager(fmu_dir)
    assert lock.is_locked() is False
    with patch("os.getpid", return_value=-1234):
        lock.acquire()

    assert lock.is_locked() is True

    with patch("os.getpid", return_value=-1234):
        lock.release()
    assert lock.is_locked() is False


def test_is_acquired_expected(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests is_acquired under expected conditions."""
    lock = LockManager(fmu_dir)
    assert lock.is_acquired() is False
    lock.acquire()
    assert lock.is_acquired() is True


def test_is_acquired_unexpected_members(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests is_acquired under expected member conditions."""
    lock = LockManager(fmu_dir)
    assert lock.is_acquired() is False
    lock.acquire()
    assert lock.is_acquired() is True
    lock_info = lock._cache
    lock._cache = None
    assert lock.is_acquired() is False
    lock._cache = lock_info
    lock._acquired_at = None
    assert lock.is_acquired() is False


@pytest.mark.parametrize(
    "is_mine, is_stale, expected",
    [
        (True, True, False),
        (True, False, True),
        (False, True, False),
        (False, False, False),
    ],
)
def test_is_acquired_unexpected_methods(
    fmu_dir: ProjectFMUDirectory,
    monkeypatch: MonkeyPatch,
    is_mine: bool,
    is_stale: bool,
    expected: bool,
) -> None:
    """Tests is_acquired under expected method conditions."""
    lock = LockManager(fmu_dir)
    assert lock.is_acquired() is False
    lock.acquire()

    with (
        patch.object(lock, "_is_mine", return_value=is_mine),
        patch.object(lock, "_is_stale", return_value=is_stale),
    ):
        assert lock.is_acquired() is expected


def test_refresh_works_as_expected(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests refresh works as expected."""
    lock = LockManager(fmu_dir)

    start_time = 1234.5
    refresh_time = 2345.6
    with patch("time.time", return_value=start_time):
        lock.acquire()
    assert lock._cache is not None
    assert lock._cache.expires_at == start_time + DEFAULT_LOCK_TIMEOUT

    with patch("time.time", return_value=refresh_time):
        lock.refresh()
    assert lock._acquired_at == start_time
    assert lock._cache.expires_at == refresh_time + DEFAULT_LOCK_TIMEOUT

    lock_info = LockInfo.model_validate(json.loads(lock.path.read_text()))
    assert lock_info.expires_at == refresh_time + DEFAULT_LOCK_TIMEOUT


def test_refresh_without_lock_file(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests refresh when lock file is not present."""
    lock = LockManager(fmu_dir)
    with pytest.raises(LockError, match="does not exist"):
        lock.refresh()

    lock.acquire()
    lock.path.unlink()  # It was deleted or something.
    with pytest.raises(LockError, match="does not exist"):
        lock.refresh()


def test_refresh_without_owning_lock(
    fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch
) -> None:
    """Tests refresh when lock file is not owned by the process."""
    lock = LockManager(fmu_dir)
    with patch("os.getpid", return_value=-1234):
        lock.acquire()
    with pytest.raises(LockError, match="held by another process"):
        lock.refresh()


def test_lock_release_unlinks_lock_file(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that releasing a lock removes the lock file."""
    lock = LockManager(fmu_dir)

    assert lock.exists is False
    lock.acquire()
    assert lock.exists
    lock.release()
    assert lock.exists is False
    assert lock._cache is None
    assert lock.path.exists() is False


def test_safe_load(fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch) -> None:
    """Tests the safe load helper method."""
    lock = LockManager(fmu_dir)
    lock.acquire()
    assert lock._cache is not None
    assert lock._safe_load() == lock._cache

    lock.release()
    lock.path.write_text("a")
    assert lock._safe_load() is None


def test_save_expected(fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch) -> None:
    """Tests that save works as expected."""
    lock = LockManager(fmu_dir)
    lock.acquire()

    lock_info = lock.load()
    new_expires_at = 123.4
    lock_info.expires_at = new_expires_at

    temp_uuid = uuid.uuid4()
    with patch("uuid.uuid4", return_value=temp_uuid):
        lock.save(lock_info)

    lock_info = lock.load()
    assert lock_info.expires_at == new_expires_at
    assert Path(f"{lock.path}.tmp.{temp_uuid.hex[:8]}").exists() is False
    assert LockInfo.model_validate(json.loads(lock.path.read_text())) == lock_info


def test_save_raises(fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch) -> None:
    """Tests that save fails as expected under failing conditions."""
    lock = LockManager(fmu_dir)
    lock.acquire()

    lock_info = lock.load()
    with (
        patch("pathlib.Path.replace", side_effect=OSError("oops")) as mock_replace,
        pytest.raises(LockError, match="oops"),
    ):
        lock.save(lock_info)
        mock_replace.assert_called_once_with(lock.path)


def test_save_not_owned(fmu_dir: ProjectFMUDirectory, monkeypatch: MonkeyPatch) -> None:
    """Tests that save fails when lock is not owned by process."""
    lock = LockManager(fmu_dir)
    lock.acquire()

    lock_info = lock.load()
    with (
        patch.object(lock, "_is_mine", return_value=False),
        pytest.raises(LockError, match="lock file is held by another"),
    ):
        lock.save(lock_info)


def test_ensure_can_write_no_lock(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests ensure_can_write when no lock exists."""
    lock = LockManager(fmu_dir)
    assert lock.exists is False
    lock.ensure_can_write()


def test_ensure_can_write_invalid_lock(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests ensure_can_write ignores unreadable lock info."""
    lock = LockManager(fmu_dir)
    lock.path.write_text("garbage")
    with patch.object(lock, "_safe_load", return_value=None):
        lock.ensure_can_write()


def test_ensure_can_write_owned_lock(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests ensure_can_write passes when lock is owned by caller."""
    lock = LockManager(fmu_dir)
    lock.acquire()
    lock.ensure_can_write()
    lock.release()


def test_ensure_can_write_stale_lock(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests ensure_can_write ignores stale locks."""
    lock = LockManager(fmu_dir)
    future = time.time() + 10
    lock_info = LockInfo(
        pid=123,
        hostname="host",
        user="user",
        acquired_at=future - 100,
        expires_at=future,
    )
    lock.path.write_text(lock_info.model_dump_json(indent=2))
    with (
        patch.object(lock, "_safe_load", return_value=lock_info),
        patch.object(lock, "is_acquired", return_value=False),
        patch.object(lock, "_is_stale", return_value=True),
    ):
        lock.ensure_can_write()


def test_ensure_can_write_foreign_lock(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests ensure_can_write raises on active locks owned by others."""
    lock = LockManager(fmu_dir)
    now = time.time()
    lock_info = LockInfo(
        pid=123,
        hostname="remote",
        user="user",
        acquired_at=now,
        expires_at=now + DEFAULT_LOCK_TIMEOUT,
    )
    lock.path.write_text(lock_info.model_dump_json(indent=2))
    with (
        patch(
            "fmu.settings._resources.lock_manager.socket.gethostname",
            return_value="current-host",
        ),
        pytest.raises(PermissionError, match="Cannot write to .fmu directory"),
    ):
        lock.ensure_can_write()



"""Main interface for working with .fmu directory."""

from pathlib import Path
from typing import Any, Final, Self, TypeAlias, cast

from ._logging import null_logger
from ._resources.config_managers import (
    ProjectConfigManager,
    UserConfigManager,
)
from ._resources.lock_manager import LockManager
from .models.project_config import ProjectConfig
from .models.user_config import UserConfig

logger: Final = null_logger(__name__)

FMUConfigManager: TypeAlias = ProjectConfigManager | UserConfigManager


class FMUDirectoryBase:
    """Provides access to a .fmu directory and operations on its contents."""

    config: FMUConfigManager
    _lock: LockManager

    def __init__(self: Self, base_path: str | Path) -> None:
        """Initializes access to a .fmu directory.

        Args:
            base_path: The directory containing the .fmu directory or one of its parent
                dirs

        Raises:
            FileExistsError: If .fmu exists but is not a directory
            FileNotFoundError: If .fmu directory doesn't exist
            PermissionError: If lacking permissions to read/write to the directory
        """
        self.base_path = Path(base_path).resolve()
        logger.debug(f"Initializing FMUDirectory from '{base_path}'")
        self._lock = LockManager(self)

        fmu_dir = self.base_path / ".fmu"
        if fmu_dir.exists():
            if fmu_dir.is_dir():
                self._path = fmu_dir
            else:
                raise FileExistsError(
                    f".fmu exists at {self.base_path} but is not a directory"
                )
        else:
            raise FileNotFoundError(f"No .fmu directory found at {self.base_path}")

        logger.debug(f"Using .fmu directory at {self._path}")

    @property
    def path(self: Self) -> Path:
        """Returns the path to the .fmu directory."""
        return self._path

    def get_config_value(self: Self, key: str, default: Any = None) -> Any:
        """Gets a configuration value by key.

        Supports dot notation for nested values (e.g., "foo.bar")

        Args:
            key: The configuration key
            default: Value to return if key is not found. Default None

        Returns:
            The configuration value or deafult
        """
        return self.config.get(key, default)

    def set_config_value(self: Self, key: str, value: Any) -> None:
        """Sets a configuration value by key.

        Args:
            key: The configuration key
            value: The value to set

        Raises:
            FileNotFoundError: If config file doesn't exist
            ValueError: If the updated config is invalid
        """
        logger.info(f"Setting {key} in {self.path}")
        self.config.set(key, value)
        logger.debug(f"Set {key} to {value}")

    def update_config(
        self: Self, updates: dict[str, Any]
    ) -> ProjectConfig | UserConfig:
        """Updates multiple configuration values at once.

        Args:
            updates: Dictionary of key-value pairs to update

        Returns:
            The updated *Config object

        Raises:
            FileNotFoundError: If config file doesn't exist
            ValueError: If the updates config is invalid
        """
        return self.config.update(updates)

    def get_file_path(self: Self, relative_path: str | Path) -> Path:
        """Gets the absolute path to a file within the .fmu directory.

        Args:
            relative_path: Path relative to the .fmu directory

        Returns:
            Absolute path to the file
        """
        return self.path / relative_path

    def read_file(self, relative_path: str | Path) -> bytes:
        """Reads a file from the .fmu directory.

        Args:
            relative_path: Path relative to the .fmu directory

        Returns:
            File contents as bytes

        Raises:
            FileNotFoundError: If the file doesn't exist
        """
        file_path = self.get_file_path(relative_path)
        return file_path.read_bytes()

    def read_text_file(self, relative_path: str | Path, encoding: str = "utf-8") -> str:
        """Reads a text file from the .fmu directory.

        Args:
            relative_path: Path relative to the .fmu directory
            encoding: Text encoding to use. Default utf-8

        Returns:
            File contents as string
        """
        file_path = self.get_file_path(relative_path)
        return file_path.read_text(encoding=encoding)

    def write_file(self, relative_path: str | Path, data: bytes) -> None:
        """Writes bytes to a file in the .fmu directory.

        Args:
            relative_path: Path relative to the .fmu directory
            data: Bytes to write
        """
        self._lock.ensure_can_write()
        file_path = self.get_file_path(relative_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)

        file_path.write_bytes(data)
        logger.debug(f"Wrote {len(data)} bytes to {file_path}")

    def write_text_file(
        self, relative_path: str | Path, content: str, encoding: str = "utf-8"
    ) -> None:
        """Writes text to a file in the .fmu directory.

        Args:
            relative_path: Path relative to the .fmu directory
            content: Text content to write
            encoding: Text encoding to use. Default utf-8
        """
        self._lock.ensure_can_write()
        file_path = self.get_file_path(relative_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)

        file_path.write_text(content, encoding=encoding)
        logger.debug(f"Wrote text file to {file_path}")

    def list_files(self, subdirectory: str | Path | None = None) -> list[Path]:
        """Lists files in the .fmu directory or a subdirectory.

        Args:
            subdirectory: Optional subdirectory to list files from

        Returns:
            List of Path objects for files (not directories)
        """
        base = self.get_file_path(subdirectory) if subdirectory else self.path
        if not base.exists():
            return []

        return [p for p in base.iterdir() if p.is_file()]

    def ensure_directory(self, relative_path: str | Path) -> Path:
        """Ensures a subdirectory exists in the .fmu directory.

        Args:
            relative_path: Path relative to the .fmu directory

        Returns:
            Path to the directory
        """
        dir_path = self.get_file_path(relative_path)
        dir_path.mkdir(parents=True, exist_ok=True)
        return dir_path

    def file_exists(self, relative_path: str | Path) -> bool:
        """Checks if a file exists in the .fmu directory.

        Args:
            relative_path: Path relative to the .fmu directory

        Returns:
            True if the file exists, False otherwise
        """
        return self.get_file_path(relative_path).exists()


class ProjectFMUDirectory(FMUDirectoryBase):
    config: ProjectConfigManager

    def __init__(self, base_path: str | Path) -> None:
        """Initializes a project-based .fmu directory."""
        self.config = ProjectConfigManager(self)
        super().__init__(base_path)

    def update_config(self: Self, updates: dict[str, Any]) -> ProjectConfig:
        """Updates multiple configuration values at once.

        Args:
            updates: Dictionary of key-value pairs to update

        Returns:
            The updated ProjectConfig object

        Raises:
            FileNotFoundError: If config file doesn't exist
            ValueError: If the updates config is invalid
        """
        return cast("ProjectConfig", super().update_config(updates))

    @staticmethod
    def find_fmu_directory(start_path: Path) -> Path | None:
        """Searches for a .fmu directory in start_path and its parents.

        Args:
            start_path: The path to start searching from

        Returns:
            Path to the found .fmu directory or None if not found
        """
        current = start_path
        # Prevent symlink loops
        visited = set()

        while current not in visited:
            visited.add(current)
            fmu_dir = current / ".fmu"

            # Do not include $HOME/.fmu in the search
            if fmu_dir.is_dir() and current != Path.home():
                return fmu_dir

            # We hit root
            if current == current.parent:
                break

            current = current.parent

        return None

    @classmethod
    def find_nearest(cls: type[Self], start_path: str | Path = ".") -> Self:
        """Factory method to find and open the nearest .fmu directory.

        Args:
            start_path: Path to start searching from. Default current working director

        Returns:
            FMUDirectory instance

        Raises:
            FileNotFoundError: If no .fmu directory is found
        """
        start_path = Path(start_path).resolve()
        fmu_dir_path = cls.find_fmu_directory(start_path)
        if fmu_dir_path is None:
            raise FileNotFoundError(f"No .fmu directory found at or above {start_path}")
        return cls(fmu_dir_path.parent)


class UserFMUDirectory(FMUDirectoryBase):
    config: UserConfigManager

    def __init__(self) -> None:
        """Initializes a project-based .fmu directory."""
        self.config = UserConfigManager(self)
        super().__init__(Path.home())

    def update_config(self: Self, updates: dict[str, Any]) -> UserConfig:
        """Updates multiple configuration values at once.

        Args:
            updates: Dictionary of key-value pairs to update

        Returns:
            The updated UserConfig object

        Raises:
            FileNotFoundError: If config file doesn't exist
            ValueError: If the updates config is invalid
        """
        return cast("UserConfig", super().update_config(updates))


def get_fmu_directory(base_path: str | Path) -> ProjectFMUDirectory:
    """Initializes access to a .fmu directory.

    Args:
        base_path: The directory containing the .fmu directory or one of its parent
                   dirs

    Returns:
        FMUDirectory instance

    Raises:
        FileExistsError: If .fmu exists but is not a directory
        FileNotFoundError: If .fmu directory doesn't exist
        PermissionError: If lacking permissions to read/write to the directory

    """
    return ProjectFMUDirectory(base_path)


def find_nearest_fmu_directory(start_path: str | Path = ".") -> ProjectFMUDirectory:
    """Factory method to find and open the nearest .fmu directory.

    Args:
        start_path: Path to start searching from. Default current working directory

    Returns:
        FMUDirectory instance

    Raises:
        FileNotFoundError: If no .fmu directory is found
    """
    return ProjectFMUDirectory.find_nearest(start_path)
